@@ -28,18 +28,18 @@ Note: as the implementation is based on FakeFilesystem, all faked classes
 (including PurePosixPath, PosixPath, PureWindowsPath and WindowsPath)
 get the properties of the underlying fake filesystem.
 """
-
+import errno
+import fnmatch
+import functools
 import os
 import pathlib
+from pathlib import PurePath
+import re
+import sys
 from urllib.parse import quote_from_bytes as urlquote_from_bytes
 
-import sys
-
-import functools
-
-import errno
-
 from pyfakefs import fake_scandir
+from pyfakefs.extra_packages import use_scandir
 from pyfakefs.fake_filesystem import FakeFileOpen, FakeFilesystem
 
 
@@ -54,8 +54,8 @@ def init_module(filesystem):
 
 def _wrap_strfunc(strfunc):
     @functools.wraps(strfunc)
-    def _wrapped(pathobj, *args):
-        return strfunc(pathobj.filesystem, str(pathobj), *args)
+    def _wrapped(pathobj, *args, **kwargs):
+        return strfunc(pathobj.filesystem, str(pathobj), *args, **kwargs)
 
     return staticmethod(_wrapped)
 
@@ -78,7 +78,13 @@ def _wrap_binary_strfunc_reverse(strfunc):
     return staticmethod(_wrapped)
 
 
-class _FakeAccessor(pathlib._Accessor):  # pylint: disable=protected-access
+try:
+    accessor = pathlib._Accessor  # type: ignore [attr-defined]
+except AttributeError:
+    accessor = object
+
+
+class _FakeAccessor(accessor):  # type: ignore [valid-type, misc]
     """Accessor which forwards some of the functions to FakeFilesystem methods.
     """
 
@@ -89,19 +95,31 @@ class _FakeAccessor(pathlib._Accessor):  # pylint: disable=protected-access
 
     listdir = _wrap_strfunc(FakeFilesystem.listdir)
 
-    chmod = _wrap_strfunc(FakeFilesystem.chmod)
-
-    if sys.version_info >= (3, 6):
+    if use_scandir:
         scandir = _wrap_strfunc(fake_scandir.scandir)
 
+    chmod = _wrap_strfunc(FakeFilesystem.chmod)
+
     if hasattr(os, "lchmod"):
         lchmod = _wrap_strfunc(lambda fs, path, mode: FakeFilesystem.chmod(
             fs, path, mode, follow_symlinks=False))
     else:
-        def lchmod(self, pathobj, mode):
+        def lchmod(self, pathobj,  *args, **kwargs):
             """Raises not implemented for Windows systems."""
             raise NotImplementedError("lchmod() not available on this system")
 
+        def chmod(self, pathobj, *args, **kwargs):
+            if "follow_symlinks" in kwargs:
+                if sys.version_info < (3, 10):
+                    raise TypeError("chmod() got an unexpected keyword "
+                                    "argument 'follow_synlinks'")
+                if (not kwargs["follow_symlinks"] and
+                        os.chmod not in os.supports_follow_symlinks):
+                    raise NotImplementedError(
+                        "`follow_symlinks` for chmod() is not available "
+                        "on this system")
+            return pathobj.filesystem.chmod(str(pathobj), *args, **kwargs)
+
     mkdir = _wrap_strfunc(FakeFilesystem.makedir)
 
     unlink = _wrap_strfunc(FakeFilesystem.remove)
@@ -111,8 +129,7 @@ class _FakeAccessor(pathlib._Accessor):  # pylint: disable=protected-access
     rename = _wrap_binary_strfunc(FakeFilesystem.rename)
 
     replace = _wrap_binary_strfunc(
-        lambda fs, old_path,
-        new_path: FakeFilesystem.rename(
+        lambda fs, old_path, new_path: FakeFilesystem.rename(
             fs, old_path, new_path, force_replace=True))
 
     symlink = _wrap_binary_strfunc_reverse(
@@ -120,13 +137,31 @@ class _FakeAccessor(pathlib._Accessor):  # pylint: disable=protected-access
         FakeFilesystem.create_symlink(fs, file_path, link_target,
                                       create_missing_dirs=False))
 
+    if (3, 8) <= sys.version_info:
+        link_to = _wrap_binary_strfunc(
+            lambda fs, file_path, link_target:
+            FakeFilesystem.link(fs, file_path, link_target))
+
+    if sys.version_info >= (3, 10):
+        link = _wrap_binary_strfunc(
+            lambda fs, file_path, link_target:
+            FakeFilesystem.link(fs, file_path, link_target))
+
+        # this will use the fake filesystem because os is patched
+        def getcwd(self):
+            return os.getcwd()
+
+    readlink = _wrap_strfunc(FakeFilesystem.readlink)
+
     utime = _wrap_strfunc(FakeFilesystem.utime)
 
 
 _fake_accessor = _FakeAccessor()
 
+flavour = pathlib._Flavour  # type: ignore [attr-defined]
 
-class _FakeFlavour(pathlib._Flavour):
+
+class _FakeFlavour(flavour):  # type: ignore [valid-type, misc]
     """Fake Flavour implementation used by PurePath and _Flavour"""
 
     filesystem = None
@@ -137,8 +172,8 @@ class _FakeFlavour(pathlib._Flavour):
     ext_namespace_prefix = '\\\\?\\'
 
     drive_letters = (
-        set(chr(x) for x in range(ord('a'), ord('z') + 1)) |
-        set(chr(x) for x in range(ord('A'), ord('Z') + 1))
+            set(chr(x) for x in range(ord('a'), ord('z') + 1)) |
+            set(chr(x) for x in range(ord('A'), ord('Z') + 1))
     )
 
     def __init__(self, filesystem):
@@ -268,6 +303,7 @@ class _FakeFlavour(pathlib._Flavour):
                     seen[newpath] = path  # resolved symlink
 
             return path
+
         # NOTE: according to POSIX, getcwd() cannot contain path components
         # which are symlinks.
         base = '' if path.is_absolute() else self.filesystem.cwd
@@ -280,13 +316,13 @@ class _FakeFlavour(pathlib._Flavour):
         previous_s = None
         if strict:
             if not self.filesystem.exists(path):
-                raise FileNotFoundError(path)
+                self.filesystem.raise_os_error(errno.ENOENT, path)
             return self.filesystem.resolve_path(path)
         else:
             while True:
                 try:
                     path = self.filesystem.resolve_path(path)
-                except FileNotFoundError:
+                except OSError:
                     previous_s = path
                     path = self.filesystem.splitpath(path)[0]
                 else:
@@ -323,9 +359,9 @@ class _FakeWindowsFlavour(_FakeFlavour):
     implementations independent of FakeFilesystem properties.
     """
     reserved_names = (
-        {'CON', 'PRN', 'AUX', 'NUL'} |
-        {'COM%d' % i for i in range(1, 10)} |
-        {'LPT%d' % i for i in range(1, 10)}
+            {'CON', 'PRN', 'AUX', 'NUL'} |
+            {'COM%d' % i for i in range(1, 10)} |
+            {'LPT%d' % i for i in range(1, 10)}
     )
 
     def is_reserved(self, parts):
@@ -381,7 +417,7 @@ class _FakeWindowsFlavour(_FakeFlavour):
             # corresponding usernames.  If current user home directory points
             # to nonstandard place, this guess is likely wrong.
             if os.environ['USERNAME'] != username:
-                drv, root, parts = self.parse_parts((userhome, ))
+                drv, root, parts = self.parse_parts((userhome,))
                 if parts[-1] != os.environ['USERNAME']:
                     raise RuntimeError("Can't determine home directory "
                                        "for %r" % username)
@@ -392,6 +428,9 @@ class _FakeWindowsFlavour(_FakeFlavour):
                     userhome = self.join(parts)
         return userhome
 
+    def compile_pattern(self, pattern):
+        return re.compile(fnmatch.translate(pattern), re.IGNORECASE).fullmatch
+
 
 class _FakePosixFlavour(_FakeFlavour):
     """Flavour used by PurePosixPath with some Unix specific implementations
@@ -423,6 +462,9 @@ class _FakePosixFlavour(_FakeFlavour):
                 raise RuntimeError("Can't determine home directory "
                                    "for %r" % username)
 
+    def compile_pattern(self, pattern):
+        return re.compile(fnmatch.translate(pattern)).fullmatch
+
 
 class FakePath(pathlib.Path):
     """Replacement for pathlib.Path. Reimplement some methods to use
@@ -437,21 +479,45 @@ class FakePath(pathlib.Path):
     def __new__(cls, *args, **kwargs):
         """Creates the correct subclass based on OS."""
         if cls is FakePathlibModule.Path:
-            cls = (FakePathlibModule.WindowsPath if os.name == 'nt'
+            cls = (FakePathlibModule.WindowsPath
+                   if cls.filesystem.is_windows_fs
                    else FakePathlibModule.PosixPath)
-        self = cls._from_parts(args, init=True)
+        self = cls._from_parts(args)
         return self
 
-    def _path(self):
-        """Returns the underlying path string as used by the fake filesystem.
-        """
-        return str(self)
+    @classmethod
+    def _from_parts(cls, args, init=False):  # pylint: disable=unused-argument
+        # Overwritten to call _init to set the fake accessor,
+        # which is not done since Python 3.10
+        self = object.__new__(cls)
+        self._init()
+        drv, root, parts = self._parse_args(args)
+        self._drv = drv
+        self._root = root
+        self._parts = parts
+        return self
+
+    @classmethod
+    def _from_parsed_parts(cls, drv, root, parts):
+        # Overwritten to call _init to set the fake accessor,
+        # which is not done since Python 3.10
+        self = object.__new__(cls)
+        self._init()
+        self._drv = drv
+        self._root = root
+        self._parts = parts
+        return self
 
     def _init(self, template=None):
         """Initializer called from base class."""
         self._accessor = _fake_accessor
         self._closed = False
 
+    def _path(self):
+        """Returns the underlying path string as used by the fake filesystem.
+        """
+        return str(self)
+
     @classmethod
     def cwd(cls):
         """Return a new path pointing to the current working directory
@@ -470,13 +536,16 @@ class FakePath(pathlib.Path):
                 New in Python 3.6.
 
         Raises:
-            IOError: if the path doesn't exist (strict=True or Python < 3.6)
+            OSError: if the path doesn't exist (strict=True or Python < 3.6)
         """
-        if strict is None:
-            strict = sys.version_info < (3, 6)
-        elif sys.version_info < (3, 6):
-            raise TypeError(
-                "resolve() got an unexpected keyword argument 'strict'")
+        if sys.version_info >= (3, 6):
+            if strict is None:
+                strict = False
+        else:
+            if strict is not None:
+                raise TypeError(
+                    "resolve() got an unexpected keyword argument 'strict'")
+            strict = True
         if self._closed:
             self._raise_closed()
         path = self._flavour.resolve(self, strict=strict)
@@ -491,7 +560,7 @@ class FakePath(pathlib.Path):
         """Open the file pointed by this path and return a fake file object.
 
         Raises:
-            IOError: if the target object is a directory, the path is invalid
+            OSError: if the target object is a directory, the path is invalid
                 or permission is denied.
         """
         if self._closed:
@@ -499,95 +568,108 @@ class FakePath(pathlib.Path):
         return FakeFileOpen(self.filesystem)(
             self._path(), mode, buffering, encoding, errors, newline)
 
-    if sys.version_info >= (3, 5):
-        def read_bytes(self):
-            """Open the fake file in bytes mode, read it, and close the file.
+    def read_bytes(self):
+        """Open the fake file in bytes mode, read it, and close the file.
 
-            Raises:
-                IOError: if the target object is a directory, the path is
-                    invalid or permission is denied.
-            """
-            with FakeFileOpen(self.filesystem)(self._path(), mode='rb') as f:
-                return f.read()
+        Raises:
+            OSError: if the target object is a directory, the path is
+                invalid or permission is denied.
+        """
+        with FakeFileOpen(self.filesystem)(self._path(), mode='rb') as f:
+            return f.read()
 
-        def read_text(self, encoding=None, errors=None):
-            """
-            Open the fake file in text mode, read it, and close the file.
-            """
-            with FakeFileOpen(self.filesystem)(self._path(), mode='r',
-                                               encoding=encoding,
-                                               errors=errors) as f:
-                return f.read()
+    def read_text(self, encoding=None, errors=None):
+        """
+        Open the fake file in text mode, read it, and close the file.
+        """
+        with FakeFileOpen(self.filesystem)(self._path(), mode='r',
+                                           encoding=encoding,
+                                           errors=errors) as f:
+            return f.read()
 
-        def write_bytes(self, data):
-            """Open the fake file in bytes mode, write to it, and close the file.
-            Args:
-                data: the bytes to be written
-            Raises:
-                IOError: if the target object is a directory, the path is
-                    invalid or permission is denied.
-            """
-            # type-check for the buffer interface before truncating the file
-            view = memoryview(data)
-            with FakeFileOpen(self.filesystem)(self._path(), mode='wb') as f:
-                return f.write(view)
+    def write_bytes(self, data):
+        """Open the fake file in bytes mode, write to it, and close the file.
+        Args:
+            data: the bytes to be written
+        Raises:
+            OSError: if the target object is a directory, the path is
+                invalid or permission is denied.
+        """
+        # type-check for the buffer interface before truncating the file
+        view = memoryview(data)
+        with FakeFileOpen(self.filesystem)(self._path(), mode='wb') as f:
+            return f.write(view)
 
-        def write_text(self, data, encoding=None, errors=None):
-            """Open the fake file in text mode, write to it, and close
-            the file.
+    def write_text(self, data, encoding=None, errors=None, newline=None):
+        """Open the fake file in text mode, write to it, and close
+        the file.
 
-            Args:
-                data: the string to be written
-                encoding: the encoding used for the string; if not given, the
-                    default locale encoding is used
-                errors: ignored
-            Raises:
-                TypeError: if data is not of type 'str'
-                IOError: if the target object is a directory, the path is
-                    invalid or permission is denied.
-            """
-            if not isinstance(data, str):
-                raise TypeError('data must be str, not %s' %
-                                data.__class__.__name__)
-            with FakeFileOpen(self.filesystem)(self._path(), mode='w',
-                                               encoding=encoding,
-                                               errors=errors) as f:
-                return f.write(data)
+        Args:
+            data: the string to be written
+            encoding: the encoding used for the string; if not given, the
+                default locale encoding is used
+            errors: (str) Defines how encoding errors are handled.
+            newline: Controls universal newlines, passed to stream object.
+                New in Python 3.10.
+        Raises:
+            TypeError: if data is not of type 'str'.
+            OSError: if the target object is a directory, the path is
+                invalid or permission is denied.
+        """
+        if not isinstance(data, str):
+            raise TypeError('data must be str, not %s' %
+                            data.__class__.__name__)
+        if newline is not None and sys.version_info < (3, 10):
+            raise TypeError("write_text() got an unexpected "
+                            "keyword argument 'newline'")
+        with FakeFileOpen(self.filesystem)(self._path(),
+                                           mode='w',
+                                           encoding=encoding,
+                                           errors=errors,
+                                           newline=newline) as f:
+            return f.write(data)
 
-        @classmethod
-        def home(cls):
-            """Return a new path pointing to the user's home directory (as
-            returned by os.path.expanduser('~')).
-            """
-            return cls(cls()._flavour.gethomedir(None).
-                       replace(os.sep, cls.filesystem.path_separator))
+    @classmethod
+    def home(cls):
+        """Return a new path pointing to the user's home directory (as
+        returned by os.path.expanduser('~')).
+        """
+        home = os.path.expanduser("~")
+        if cls.filesystem.is_windows_fs != (os.name == 'nt'):
+            username = os.path.split(home)[1]
+            if cls.filesystem.is_windows_fs:
+                home = os.path.join('C:', 'Users', username)
+            else:
+                home = os.path.join('home', username)
+            cls.filesystem.create_dir(home)
+        return cls(home.replace(os.sep, cls.filesystem.path_separator))
 
-        def samefile(self, other_path):
-            """Return whether other_path is the same or not as this file
-            (as returned by os.path.samefile()).
+    def samefile(self, other_path):
+        """Return whether other_path is the same or not as this file
+        (as returned by os.path.samefile()).
 
-            Args:
-                other_path: A path object or string of the file object
-                to be compared with
+        Args:
+            other_path: A path object or string of the file object
+            to be compared with
 
-            Raises:
-                OSError: if the filesystem object doesn't exist.
-            """
-            st = self.stat()
-            try:
-                other_st = other_path.stat()
-            except AttributeError:
-                other_st = self.filesystem.stat(other_path)
-            return (st.st_ino == other_st.st_ino and
-                    st.st_dev == other_st.st_dev)
+        Raises:
+            OSError: if the filesystem object doesn't exist.
+        """
+        st = self.stat()
+        try:
+            other_st = other_path.stat()
+        except AttributeError:
+            other_st = self.filesystem.stat(other_path)
+        return (st.st_ino == other_st.st_ino and
+                st.st_dev == other_st.st_dev)
 
-        def expanduser(self):
-            """ Return a new path with expanded ~ and ~user constructs
-            (as returned by os.path.expanduser)
-            """
-            return FakePath(os.path.expanduser(self._path())
-                            .replace(os.path.sep,
-                                     self.filesystem.path_separator))
+    def expanduser(self):
+        """ Return a new path with expanded ~ and ~user constructs
+        (as returned by os.path.expanduser)
+        """
+        return FakePath(os.path.expanduser(self._path())
+                        .replace(os.path.sep,
+                                 self.filesystem.path_separator))
 
     def touch(self, mode=0o666, exist_ok=True):
         """Create a fake file for the path with the given access mode,
@@ -599,23 +681,25 @@ class FakePath(pathlib.Path):
                 happens, otherwise FileExistError is raised
 
         Raises:
-            FileExistsError if the file exists and exits_ok is False.
+            FileExistsError: if the file exists and exits_ok is False.
         """
         if self._closed:
             self._raise_closed()
         if self.exists():
             if exist_ok:
-                self.filesystem.utime(self._path(), None)
+                self.filesystem.utime(self._path(), times=None)
             else:
-                raise FileExistsError
+                self.filesystem.raise_os_error(errno.EEXIST, self._path())
         else:
             fake_file = self.open('w')
             fake_file.close()
             self.chmod(mode)
 
 
-class FakePathlibModule(object):
+class FakePathlibModule:
     """Uses FakeFilesystem to provide a fake pathlib module replacement.
+    Can be used to replace both the standard `pathlib` module and the
+    `pathlib2` package available on PyPi.
 
     You need a fake_filesystem to use this:
     `filesystem = fake_filesystem.FakeFilesystem()`
@@ -632,30 +716,141 @@ class FakePathlibModule(object):
         init_module(filesystem)
         self._pathlib_module = pathlib
 
-    class PurePosixPath(pathlib.PurePath):
+    class PurePosixPath(PurePath):
         """A subclass of PurePath, that represents non-Windows filesystem
         paths"""
         __slots__ = ()
 
-    class PureWindowsPath(pathlib.PurePath):
+    class PureWindowsPath(PurePath):
         """A subclass of PurePath, that represents Windows filesystem paths"""
         __slots__ = ()
 
-    if sys.platform == 'win32':
-        class WindowsPath(FakePath, PureWindowsPath):
-            """A subclass of Path and PureWindowsPath that represents
-            concrete Windows filesystem paths.
+    class WindowsPath(FakePath, PureWindowsPath):
+        """A subclass of Path and PureWindowsPath that represents
+        concrete Windows filesystem paths.
+        """
+        __slots__ = ()
+
+        def owner(self):
+            raise NotImplementedError(
+                "Path.owner() is unsupported on this system")
+
+        def group(self):
+            raise NotImplementedError(
+                "Path.group() is unsupported on this system")
+
+        def is_mount(self):
+            raise NotImplementedError(
+                "Path.is_mount() is unsupported on this system")
+
+    class PosixPath(FakePath, PurePosixPath):
+        """A subclass of Path and PurePosixPath that represents
+        concrete non-Windows filesystem paths.
+        """
+        __slots__ = ()
+
+        def owner(self):
+            """Return the current user name. It is assumed that the fake
+            file system was created by the current user.
             """
-            __slots__ = ()
-    else:
-        class PosixPath(FakePath, PurePosixPath):
-            """A subclass of Path and PurePosixPath that represents
-            concrete non-Windows filesystem paths.
+            import pwd
+
+            return pwd.getpwuid(os.getuid()).pw_name
+
+        def group(self):
+            """Return the current group name. It is assumed that the fake
+            file system was created by the current user.
             """
-            __slots__ = ()
+            import grp
+
+            return grp.getgrgid(os.getgid()).gr_name
 
     Path = FakePath
 
     def __getattr__(self, name):
         """Forwards any unfaked calls to the standard pathlib module."""
         return getattr(self._pathlib_module, name)
+
+
+class FakePathlibPathModule:
+    """Patches `pathlib.Path` by passing all calls to FakePathlibModule."""
+    fake_pathlib = None
+
+    def __init__(self, filesystem=None):
+        if self.fake_pathlib is None:
+            self.__class__.fake_pathlib = FakePathlibModule(filesystem)
+
+    def __call__(self, *args, **kwargs):
+        return self.fake_pathlib.Path(*args, **kwargs)
+
+    def __getattr__(self, name):
+        return getattr(self.fake_pathlib.Path, name)
+
+
+class RealPath(pathlib.Path):
+    """Replacement for `pathlib.Path` if it shall not be faked.
+    Needed because `Path` in `pathlib` is always faked, even if `pathlib`
+    itself is not.
+    """
+
+    def __new__(cls, *args, **kwargs):
+        """Creates the correct subclass based on OS."""
+        if cls is RealPathlibModule.Path:
+            cls = (RealPathlibModule.WindowsPath if os.name == 'nt'
+                   else RealPathlibModule.PosixPath)
+        self = cls._from_parts(args)
+        return self
+
+
+class RealPathlibModule:
+    """Used to replace `pathlib` for skipped modules.
+    As the original `pathlib` is always patched to use the fake path,
+    we need to provide a version which does not do this.
+    """
+
+    def __init__(self):
+        RealPathlibModule.PureWindowsPath._flavour = pathlib._WindowsFlavour()
+        RealPathlibModule.PurePosixPath._flavour = pathlib._PosixFlavour()
+        self._pathlib_module = pathlib
+
+    class PurePosixPath(PurePath):
+        """A subclass of PurePath, that represents Posix filesystem paths"""
+        __slots__ = ()
+
+    class PureWindowsPath(PurePath):
+        """A subclass of PurePath, that represents Windows filesystem paths"""
+        __slots__ = ()
+
+    if sys.platform == 'win32':
+        class WindowsPath(RealPath, PureWindowsPath):
+            """A subclass of Path and PureWindowsPath that represents
+            concrete Windows filesystem paths.
+            """
+            __slots__ = ()
+    else:
+        class PosixPath(RealPath, PurePosixPath):
+            """A subclass of Path and PurePosixPath that represents
+            concrete non-Windows filesystem paths.
+            """
+            __slots__ = ()
+
+    Path = RealPath
+
+    def __getattr__(self, name):
+        """Forwards any unfaked calls to the standard pathlib module."""
+        return getattr(self._pathlib_module, name)
+
+
+class RealPathlibPathModule:
+    """Patches `pathlib.Path` by passing all calls to RealPathlibModule."""
+    real_pathlib = None
+
+    def __init__(self):
+        if self.real_pathlib is None:
+            self.__class__.real_pathlib = RealPathlibModule()
+
+    def __call__(self, *args, **kwargs):
+        return self.real_pathlib.Path(*args, **kwargs)
+
+    def __getattr__(self, name):
+        return getattr(self.real_pathlib.Path, name)
